@@ -4,6 +4,7 @@
 """
 
 from __future__ import print_function
+import datetime as datetime
 
 __copyright__ = """
 Copyright (C) 2005, Chuck Lever <cel@netapp.com>
@@ -225,7 +226,12 @@ Nfsv4ops = [
     'ALLOCATE',
     'DEALLOCATE',
     'LAYOUTSTATS',
-    'CLONE'
+    'CLONE',
+    'COPY',
+    'OFFLOAD_CANCEL',
+    'LOOKUPP',
+    'LAYOUTERROR',
+    'COPY_NOTIFY'
 ]
 
 class DeviceData:
@@ -386,6 +392,7 @@ class DeviceData:
         """Pretty-print the NFS options
         """
         print('  NFS mount options: %s' % ','.join(self.__nfs_data['mountoptions']))
+        print('  NFS mount age: %s' % datetime.timedelta(seconds = self.__nfs_data['age']))
         print('  NFS server capabilities: %s' % ','.join(self.__nfs_data['servercapabilities']))
         if 'nfsv4flags' in self.__nfs_data:
             print('  NFSv4 capability flags: %s' % ','.join(self.__nfs_data['nfsv4flags']))
